/** @file
 * @brief VRP buffers (frame) management
 * @ingroup VRP
 */

#include "VRP-internals.h"



/* ******************************************************************
 * **            VRP buffers
 * ******************************************************************
 */


vrp_buffer_t vrp_buffer_construct(char*data, int size)
{
  vrp_buffer_t buffer = vrp_malloc(sizeof(struct vrp_buffer_s));
  buffer->data = data;
  buffer->size = size;
  buffer->critical = NULL;
  buffer->max_loss_rate = -1;
  buffer->max_cons_loss = -1;
  return buffer;
}

void vrp_buffer_destroy(vrp_buffer_t buffer)
{
  struct vrp_critical_cell_s*cs;

  vrp_enter();
  cs = buffer->critical;
  while(cs != NULL)
    {
      buffer->critical = cs->next;
      free(cs);
      cs = buffer->critical;
    }
  buffer->critical = NULL;
  vrp_free(buffer);
  vrp_exit();
}

void vrp_buffer_add_critical(vrp_buffer_t buffer, char*data, int size)
{
  struct vrp_critical_cell_s*cell;

  if((data < buffer->data) || (data > buffer->data + buffer->size))
    {
      vrp_warning("VRP: given critical interval outside data");
    }
  else
    {
      cell = (struct vrp_critical_cell_s*) malloc(sizeof(struct vrp_critical_cell_s));
      cell->next = buffer->critical;
      buffer->critical = cell;
      cell->begin = data - buffer->data;
      cell->end   = cell->begin + size;
    }
}

void vrp_buffer_set_loss_rate(vrp_buffer_t buffer, int max, int cons)
{
  printf("setting loss rate to %d%% and %d consecutive byte \n",max,cons);
  buffer->max_loss_rate = max * 256 / 100;  
  buffer->max_cons_loss = cons;
}

