#!/bin/sh

# PM2: Parallel Multithreaded Machine
# Copyright (C) 2001 "the PM2 team" (see AUTHORS file)
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.


PROGNAME=pm2-gen-make.sh

#############################################################
# Setting destination directory

make_dir() {
    ldir=$1

    if [ ! -d $ldir ]; then
#	echo "Creating directory : $ldir"
	mkdir -p $ldir
    fi
}

############################################################
# Generating Makefiles

set_stdout() {
    dst="$1"
    if [ "$in_file" ]; then
	[ "$display" = yes ] && echo Generating "$dst" 1>&2
	set > "$dst"
    else
	[ "$display" = yes ] && echo Generating `basename "$dst"` 1>&2
    fi
}

erreur () {
    while [ $# != 0 ]; do
	echo "$1" 1>&2
	shift
    done
    exit 1
}

erreur_gen () {
    dst="$1"
    shift
    set_stdout "$dst"
    while [ $# != 0 ]; do
	echo "$1" 1>&2
	echo '$(warning '"$1"' in this flavor !)'
	shift
    done
    exit 0
}

## Main ########################################

gen_main() { 
    set_stdout "$1"

    echo "# Do not edit ! Generated by $PROGNAME"
    echo ""
    echo "MAIN_STAMP_FLAVOR := $_PM2CONFIG_FLAVOR_STAMP_FILE"
    echo "LIBS := $LIBRARIES"
    echo "PROF_LIBS := $_PM2CONFIG_PROF_MODULES"
    echo "DOT_H_GEN_LIBS := $_PM2CONFIG_HEADGEN_MODULES"
    case " $_PM2CONFIG_CFLAGS " in
	*\ -DAPPLI_PROFILE\ *)
	    echo "APP_PROFILE := yes"
	    ;;
	*)
	    echo "APP_PROFILE := no"
	    ;;
    esac
}

## Modules (libs+progs) ########################################

gen_module() {
    MODULE=$1

    eval _PM2CONFIG_MAKEFILE='$'_PM2CONFIG_${MODULE}_MAKEFILE
    eval _PM2CONFIG_LIB_DIR='$'_PM2CONFIG_${MODULE}_LIB_DIR
    eval _PM2CONFIG_OBJ_DIR='$'_PM2CONFIG_${MODULE}_OBJ_DIR
    eval _PM2CONFIG_CPP_DIR='$'_PM2CONFIG_${MODULE}_CPP_DIR
    eval _PM2CONFIG_ASM_DIR='$'_PM2CONFIG_${MODULE}_ASM_DIR
    eval _PM2CONFIG_DEP_DIR='$'_PM2CONFIG_${MODULE}_DEP_DIR
    eval _PM2CONFIG_BIN_DIR='$'_PM2CONFIG_${MODULE}_BIN_DIR
    eval _PM2CONFIG_TMP_DIR='$'_PM2CONFIG_${MODULE}_TMP_DIR
    eval _PM2CONFIG_SRC_DIR='$'_PM2CONFIG_${MODULE}_SRC_DIR
    eval _PM2CONFIG_INCLUDE_DIR='$'_PM2CONFIG_${MODULE}_INCLUDE_DIR
    # Nothing to do for stampdir...
    # Nothing to do for (flavor) stampfile
    # Nothing to do for extension
    eval _PM2CONFIG_KERNEL_CFLAGS='$'_PM2CONFIG_${MODULE}_KERNEL_CFLAGS
    # Nothing to do for CC
   
cat <<EOF
# Do not edit ! Generated by $PROGNAME

$_PM2CONFIG_MAKEFILE

MOD_GEN_LIB := $_PM2CONFIG_LIB_DIR
MOD_GEN_OBJ := $_PM2CONFIG_OBJ_DIR
MOD_GEN_CPP := $_PM2CONFIG_CPP_DIR
MOD_GEN_ASM := $_PM2CONFIG_ASM_DIR
MOD_GEN_DEP := $_PM2CONFIG_DEP_DIR
MOD_GEN_BIN := $_PM2CONFIG_BIN_DIR
MOD_GEN_TMP := $_PM2CONFIG_TMP_DIR
MOD_GEN_SRC := $_PM2CONFIG_SRC_DIR
MOD_GEN_INC := $_PM2CONFIG_INCLUDE_DIR
MOD_GEN_STAMP := $_PM2CONFIG_STAMP_DIR
MOD_STAMP_FLAVOR := $_PM2CONFIG_FLAVOR_STAMP_FILE
MOD_EXT := $_PM2CONFIG_EXT

MOD_CFLAGS := $_PM2CONFIG_KERNEL_CFLAGS
CFLAGS += \$(MOD_CFLAGS) -I\$(MOD_GEN_INC)
                        #^^^^^^^^^^^^^^^ should be included in pm2-config

CC := $_PM2CONFIG_CC

EOF
}

## Libs ########################################

gen_libs() {

    LIBRARY=$1
    dst=$2
    
    set_stdout "$dst"

    gen_module $LIBRARY

    eval _PM2CONFIG_BUILD_DYNAMIC='$'_PM2CONFIG_${LIBRARY}_BUILD_DYNAMIC
    eval _PM2CONFIG_BUILD_STATIC='$'_PM2CONFIG_${LIBRARY}_BUILD_STATIC
    eval _PM2CONFIG_LINK_DYNAMIC='$'_PM2CONFIG_${LIBRARY}_LINK_DYNAMIC
    eval _PM2CONFIG_LIB_PM2_SHLIBS='$'_PM2CONFIG_${LIBRARY}_PM2_SHLIBS

    cat <<EOF

# Library specific

BUILD_DYNAMIC_LIBS := $_PM2CONFIG_BUILD_DYNAMIC
BUILD_STATIC_LIBS := $_PM2CONFIG_BUILD_STATIC
LINK_DYNAMIC_LIBS := $_PM2CONFIG_LINK_DYNAMIC

MOD_PM2_SHLIBS := $_PM2CONFIG_LIB_PM2_SHLIBS
#LINK_LD_FLAGS := $_PM2CONFIG_LIBS

EOF

}

## Progs ########################################

gen_progs() {

    PROGRAM="$1"
    dst=$2

    set_stdout "$2"

    if [ -z "$PROGRAM" ]; then
	echo "Error : variable PROGRAM not set !"
	exit -1
    fi

    gen_module $PROGRAM

    cat <<EOF

# Program specific

MOD_STAMP_FILES := $_PM2CONFIG_LIB_STAMP_FILES

MOD_LDFLAGS := $_PM2CONFIG_LIBS
LDFLAGS += \$(MOD_LDFLAGS)

MOD_LLIBS += $_PM2CONFIG_LD_LIBS

# Necessaire ?
MOD_BUILDDIR :=  $_PM2CONFIG_BUILD

EOF
}

## Apps ########################################

gen_apps() {
    subdir="$1"
    dst="$2"
    set_stdout "$dst"

    cat <<EOF
# Do not edit ! Generated by $PROGNAME

MOD_GEN_OBJ := $_PM2CONFIG_BUILD/$subdir/obj
MOD_GEN_CPP := $_PM2CONFIG_BUILD/$subdir/cpp
MOD_GEN_ASM := $_PM2CONFIG_BUILD/$subdir/asm
MOD_GEN_DEP := $_PM2CONFIG_BUILD/$subdir/dep
MOD_GEN_BIN := $_PM2CONFIG_BUILD/$subdir/bin
MOD_GEN_TMP := $_PM2CONFIG_BUILD/$subdir/tmp
MOD_GEN_SRC := $_PM2CONFIG_BUILD/$subdir/src
MOD_GEN_INC := $_PM2CONFIG_BUILD/$subdir/include
MOD_GEN_STAMP := $_PM2CONFIG_STAMP_DIR
MOD_STAMP_FLAVOR := $_PM2CONFIG_FLAVOR_STAMP_FILE
MOD_STAMP_FILES := $_PM2CONFIG_LIB_STAMP_FILES
MOD_EXT := $_PM2CONFIG_EXT

MOD_BUILDD := $_PM2CONFIG_BUILD
MOD_LOADER := $_PM2CONFIG_LOADER
MOD_CFLAGS += $_PM2CONFIG_appli_KERNEL_CFLAGS
MOD_LDFLAGS += $_PM2CONFIG_LIBS
MOD_LLIBS += $_PM2CONFIG_LD_LIBS

CFLAGS += \$(MOD_CFLAGS) -I\$(MOD_GEN_INC)
                        #^^^^^^^^^^^^^^^ should be included in pm2-config
LDFLAGS += \$(MOD_LDFLAGS)
CC := $_PM2CONFIG_CC


EOF

    case " $_PM2CONFIG_CFLAGS " in
	*\ -DAPPLI_PROFILE\ *)
            echo "APP_PROFILE := yes" ;;
	*)
	    echo "APP_PROFILE := no" ;;
    esac
}

############################################################
# Main

display=yes
PM2_FLAVOR=""
flavor_set=""
modules_libs=""
modules_progs=""
modules_apps=""
modules_misc=""
modules_set=""
create_dir=""
in_file=""
which="misc"
subdir=""

while [ $# != 0 ]; do
    case "$1" in
	--type)
	    shift
	    case "$1" in
		lib|app|prog)
		    which=${1}s;;
		libs|apps|progs|misc)
		    which=${1};;
		*)
		    erreur "type $1 inconnu";;
	    esac;;
	--silent)
	    display="";;
	--create-dir)
	    create_dir=yes;;
	--stdout)
	    in_file="";;
	--in-file)
	    in_file=yes;;
	--flavor)
	    shift
	    PM2_FLAVOR="$1"
	    flavor_set=yes;;
	--subdir)
	    shift
	    subdir="$1";;
	-*)
	    erreur "Unknown option $1";;
	"")
	    erreur "Invalid (null) argument";;
	*)
	    if [ "$flavor_set" ]; then
		eval modules_${which}='"$modules_'${which}' $1"'
		modules_set="$modules_set yes"
	    else
		PM2_FLAVOR="$1"
		flavor_set=yes
	    fi;;
    esac
    shift
done

if [ -z "$flavor_set" ]; then
    erreur "no flavor set"
fi

if [ -z "$modules_set" ]; then
    modules_misc=all
fi

set -- --source-mode
. ${PM2_ROOT}/bin/pm2-config

# Generating directories

dir=$_PM2CONFIG_BUILD
LIBRARIES="$_PM2CONFIG_MODULES"
destdir=$_PM2CONFIG_MAK_DIR

# Generating .mak for libraries

test "$in_file" && make_dir $destdir

case " $modules_misc " in
    *" all "*)
	modules_misc="$modules_misc main"
	modules_libs="$modules_libs $LIBRARIES"
	;;
esac

for module in $modules_misc ; do
    case "$module" in
	all) ;;
	main|main-config)
	    gen_main "$destdir/main-config.mak" ;;
	app|apps)
	    gen_apps examples "$destdir/apps-config.mak" ;;
	prog|progs)
	    gen_progs "$PROGRAM" "$destdir/progs-config.mak" ;;
	dir)
	    ;;
	*)
	    case " $LIBRARIES " in
		*" $module "*)
		    gen_libs "$module" "$destdir/$module-config.mak";;
		*)
		    erreur_gen "$destdir/$module-config.mak" \
			"Unknown module $module";;
	    esac;;
    esac
done

for module in $modules_libs; do
    case " $LIBRARIES " in
	*" $module "*)
	    gen_libs "$module" "$destdir/$module-config.mak";;
	*)
	    erreur_gen "$destdir/$module-config.mak" \
		"Unknown library module $module";;
    esac
done

for module in $modules_progs; do
    case " $LIBRARIES " in
	*" $module "*)
	    gen_progs "$module" "$destdir/$module-config.mak"
	    ;;
	*)
	    erreur_gen "$destdir/$module-config.mak" \
		"Unknown program module $module"
	    ;;
    esac
done
    
for module in $modules_apps; do
    gen_apps "${subdir:-examples}" "$destdir/$module-config.mak"
done

exit 0
